// URL_Window.cpp

#define DEBUG 0
#include <Debug.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "URL_Window.h"

URL_Window::URL_Window(BFile * a_settings_file)
: BWindow(BRect(300,200,700,464), "URL Handlers", B_TITLED_WINDOW, B_NOT_ZOOMABLE)
{
	//	init pointers to objects
	m_settings_file		=	a_settings_file;
	m_mime_type			=	new	BMimeType();
	
	// m_topbox
	m_topbox	=	new BBox(BRect(1,1,Bounds().right,Bounds().bottom), "view", B_FOLLOW_ALL_SIDES, B_NAVIGABLE);
	AddChild(m_topbox);

	// m_listview
	m_listview = new BListView(BRect(11,11,150,m_topbox->Bounds().bottom -11), 
		"List", B_SINGLE_SELECTION_LIST, B_FOLLOW_ALL_SIDES);
	m_listview->SetSelectionMessage(new BMessage('slct'));
	m_topbox	->AddChild(new BScrollView("Scroll", m_listview, B_FOLLOW_ALL_SIDES, B_PLAIN_BORDER, false, true));

	// m_details_box
	m_details_box	=	new BBox(BRect(
							m_topbox->Bounds().right - 222,
							m_topbox->Bounds().top + 8,
							m_topbox->Bounds().right - 9,
							m_topbox->Bounds().bottom - 9), 
		"description", B_FOLLOW_TOP_BOTTOM | B_FOLLOW_RIGHT);
	m_topbox->	AddChild(m_details_box);

	// m_pref_app_menu
	m_pref_app_menu	= new BMenu("None");
	m_pref_app_menu->SetRadioMode(true);
	m_pref_app_menu->SetLabelFromMarked(true);
	
	// m_menufield
	m_menufield	=	new BMenuField(BRect(10,120,200,150), "app dropdown", NULL, m_pref_app_menu);
	BuildAppDropdown();
	m_details_box->	AddChild(m_menufield);

	// pref_app_label
	BStringView	* prefapp_label = new BStringView(BRect(10,90,150,120), "prefapp", "Preferred Application",
		B_FOLLOW_RIGHT | B_FOLLOW_TOP);
	prefapp_label->SetFont(be_bold_font);
	m_details_box->	AddChild(prefapp_label);
	
	// description_label
	BStringView	* description_label = new BStringView(BRect(10,2,150,30), "description_label", "Description",
		B_FOLLOW_RIGHT | B_FOLLOW_TOP);
	description_label->SetFont(be_bold_font);
	m_details_box->	AddChild(description_label);

	// m_longdescription
	m_longdescription = new BTextView(BRect(10,30,200,80), "description", BRect(0,0,190,50),	
		B_FOLLOW_RIGHT | B_FOLLOW_TOP);
	m_longdescription->SetViewColor(216, 216, 216);
	m_longdescription->MakeEditable(false);
	m_longdescription->MakeSelectable(false);
	m_details_box->	AddChild(m_longdescription);
	
	// interface, finishing touches
	SetSizeLimits(350, 65000, 235, 65000);
	ResizeTo(350,264);

	ReadSettings();
}

URL_Window::~URL_Window(void)	
{
	m_mime_type->StopWatching(*m_win_messenger);
	delete m_win_messenger;
	delete m_mime_type;
}

void 
URL_Window::MessageReceived(BMessage * a_message)
{
	switch(a_message->what)
	{
		case 'init':											// at startup
				DelayedInit();
				break;

		case 'slct':											// listview changed
				BuildAppDropdown();
				DisplayDescriptions();
				break;
	
		case 'papp':											// pref app selected
				SetPrefApp(a_message);
				break;

		case B_META_MIME_CHANGED:								// metamime db changed
				DealWithTypeChanges(a_message);
				break;

		default:
				BWindow::MessageReceived(a_message);
				break;			
	}
}

void
URL_Window::DelayedInit(void)
{
	// DelayedInit() makes the window appear noticeably 
	// quicker, since the mimetype lookup in PopulateList()
	// is done -after- the window is put on-screen, 
	// rather than before.
	
	PopulateList();

	// subscribe to metamime db updates		
	m_win_messenger	=	new BMessenger(NULL, this);
	m_mime_type->StartWatching(*m_win_messenger);

	m_listview->MakeFocus();
	m_listview->Select(0);
}
void
URL_Window::PrintErrorMessage(const char * a_text, int32 a_status)
{
	PRINT(("%s: %s\n", a_text, strerror(a_status)));
}

void
URL_Window::Quit(void)
{
	SaveSettings();
	be_app->PostMessage(B_QUIT_REQUESTED);
	BWindow::Quit();
}

void
URL_Window::SaveSettings(void)
{
	status_t	status	=	B_OK;
	
	BMessage msg('sett');
	msg.AddRect("winframe", Frame());
	
	m_settings_file->	Seek(0, SEEK_SET);
	
	if (m_settings_file->Position() != 0)
		PrintErrorMessage("SaveSettings(), Seek(): ", status);
	
	status	=	msg.Flatten(m_settings_file);
	if (status != B_OK)
		PrintErrorMessage("SaveSettings(), Flatten()", status);
}

void
URL_Window::ReadSettings(void)
{
	status_t	status;
	BMessage	settings;
	BRect		winframe;
	
	status	=	settings.Unflatten(m_settings_file);
	
	if (status != B_OK)
		PrintErrorMessage("Unflatten()", status);
	
	if (settings.FindRect("winframe", &winframe) == B_OK)
	{
		MoveTo(winframe.LeftTop());
		ResizeTo(winframe.Width(), winframe.Height());
	}
	else
		CenterWindowOnScreen();

	PushWindowOnScreen();
}

void
URL_Window::PopulateList(void)
{
	BMessage types_msg;
	
	m_mime_type->GetInstalledTypes("application", & types_msg);
	
	BString		type_string;
	int32		count	=	0;

	while(types_msg.FindString("types", count++, &type_string) == B_OK)
	{
		if (type_string.IFindFirst("application/x-vnd.Be.URL.") == 0)
		{
			PRINT(("type: %s\n", type_string.String()));
			
			type_string.IReplaceAll("application/x-vnd.Be.URL.", "");
			//type_string.Capitalize();
			
			m_listview->AddItem(new BStringItem(type_string.String()));
		}
	}
}

void
URL_Window::BuildAppDropdown(void)
{
	BMimeType	mime;
	BString	*	type = GetCurrentType();
	mime.SetTo(type->String());

	// Create basic app dropdown
	EmptyAppDropdown();
	BMessage * none_msg = new BMessage('papp');
	none_msg->AddString("type", type->String());
	none_msg->AddString("pref_app", "none");
	BMenuItem *	item	=	new BMenuItem("None", none_msg);
	item->SetMarked(true);
	m_pref_app_menu->AddItem(item);
	m_pref_app_menu->AddSeparatorItem();

	if ((*type) == "none")
	{	
		delete type;
		return;
	}
	
	// Get Prefered App
	status_t	status					=	B_OK;
	bool		got_pref_app_sig		=	false;
	bool		found_pref_app			=	false;
	entry_ref	pref_app_ref;
	char *		pref_app_sig 			=	new char [B_MIME_TYPE_LENGTH];

	status 	=	mime.GetPreferredApp(pref_app_sig);
	BString		pref_app_sig_string;

	if (pref_app_sig[0] == '\0')
		PRINT(("pref_app_sig == NULL\n"));
	
	if (status == B_OK)
	{
		got_pref_app_sig	=	true;
		pref_app_sig_string	=	pref_app_sig;
		pref_app_sig_string.ToLower();
	}
	else
		PrintErrorMessage("mime.GetPreferredApp(): ", status);

	delete [] pref_app_sig;

	// Get Supporting Apps
	BMessage	app_msg;
	mime.GetSupportingApps(& app_msg);

	BString		app_string;
	int32		count	=	0;

	printf("SupportingApps\n");
	app_msg.PrintToStream();

	while(app_msg.FindString("applications", count++, &app_string) == B_OK)
	{
		PRINT(("type: %s\n", app_string.String()));
	
		BList *	ref_list	=	NULL;
	
		ref_list = GetAppRefsByQuery(app_string.String());
	
		int32 app_count	=	ref_list->CountItems();
		
		if (app_count == 1)
		{
			entry_ref * ref = (entry_ref *) ref_list->FirstItem();
			
			BMessage * msg = new BMessage('papp'); // fill with info
			msg->AddString("type", type->String());
			msg->AddString("pref_app", app_string);
			msg->AddRef("app_hint", ref);
			
			URL_MenuItem *	menuitem = new URL_MenuItem(ref->name, msg, 0, 0, ref);
			
			// is this the preferred application ?
			app_string.ToLower();
			
			if (app_string == pref_app_sig_string)
			{
				found_pref_app		=	true;
				menuitem->SetMarked(true);
			}
		
			m_pref_app_menu->AddItem(menuitem);
		}
		
		if (app_count > 1)
		{
			PRINT(("app_count > 1\n"));

			// add apps
			for (int32 i = 0; i < app_count; i++)
			{
				entry_ref * ref = (entry_ref *) ref_list->ItemAt(i);
				BPath	path(ref);
				
				BMessage * msg = new BMessage('papp');
				msg->AddString("type", type->String());
				msg->AddString("pref_app", app_string);
				msg->AddRef("app_hint", ref);

				URL_MenuItem *	menuitem = new URL_MenuItem(path.Path(), msg, 0, 0, ref);

				// is this the preferred application ?
				app_string.ToLower();

				m_pref_app_menu->AddItem(menuitem); 
				
				if (app_string == pref_app_sig_string)
				{
					found_pref_app	=	true;
	
					entry_ref	hint_ref;
					mime.GetAppHint(& hint_ref);
					if (hint_ref == (*ref))			
						menuitem->SetMarked(true);
				}
			}
		}

		delete ref_list;
	}
	
	// Mark 'None', if necessary
	if ((! got_pref_app_sig) || (! found_pref_app))
	{
		BMenuItem *	menuitem;
		menuitem	=	m_pref_app_menu->ItemAt(0);
		menuitem->SetMarked(true);
		m_menufield->SetLabel(menuitem->Label());
/*		
		BMessage * msg = new BMessage('papp'); // fill with info
		msg->AddString("type", type->String());
		msg->AddString("pref_app", "none");
		menuitem->SetMessage(msg);
*/
	}
	
	delete type;
}

BList * 
URL_Window::GetAppRefsByQuery(const char * a_app_sig)
{
	BList * ref_list = new BList;
	BVolumeRoster	roster;
	BVolume			volume;
	BQuery			query;
	BString			query_predicate	("BEOS:APP_SIG='");
	entry_ref		ref;
	
	query_predicate.Append(a_app_sig);
	query_predicate.Append("'&&BEOS:TYPE='application/x-vnd.Be-elfexecutable'");
	
	while (roster.GetNextVolume(& volume) == B_OK)
	{
		if (volume.KnowsAttr() && volume.KnowsMime() && volume.KnowsQuery())
		{
			query.Clear();
			query.SetVolume(& volume);
			query.SetPredicate(query_predicate.String());
			query.Fetch();

			while (query.GetNextRef(& ref) == B_OK)
			{
				PRINT(("app: %s\n", ref.name));
				entry_ref * app_ref = new entry_ref(ref);
				ref_list->AddItem(app_ref);
			}
		}
	}
	
	return ref_list;
}

void
URL_Window::DealWithTypeChanges(BMessage * a_message)
{
	PRINT(("B_META_MIME_CHANGED\n"));

	int32		which;
	BString		type;
	BString		extra_type;

	a_message->	FindInt32("be:which", & which);
	a_message->	FindString("be:type", & type);
	a_message->	FindString("be:extra_type", & extra_type);

	PRINT(("type: %s\n", type.String()));
	PRINT(("extra_type: %s\n", extra_type.String()));
	PRINT(("which: %ld\n", which));

	type.ToLower();

	if (which & B_PREFERRED_APP_CHANGED)
	{
		BString * displayed_type = GetCurrentType();
	
		if	(displayed_type->IFindFirst(type) == 0)
			BuildAppDropdown();
			
		delete displayed_type;
	}
	
	// meta mime created - unofficial which value ?
	if (which & 256)
	{
		EmptyAppDropdown();
		EmptyListview();
		PopulateList();
	}
		
	// meta mime removed - unofficial which value ?
	if (which & 512)		
	{
		EmptyAppDropdown();
		EmptyListview();
		PopulateList();	
	}		
}

BString * 
URL_Window::GetCurrentType(void)
{
	BString * type = new BString ("application/x-vnd.Be.URL.");
	
	// get current listview item
	BStringItem * stringitem = (BStringItem *) m_listview->ItemAt(m_listview->CurrentSelection());

	if (stringitem <= 0)
	{
		type->SetTo("none");
		return type;
	}

	type->Append(stringitem->Text());
	
	return type;
}

void 
URL_Window::SetPrefApp(BMessage * a_message)
{
	BString		type;
	BString		pref_app;
	entry_ref	app_hint_ref;
	BEntry		app_hint_entry;
	
	a_message->	FindString("type", & type);
	a_message->	FindString("pref_app", & pref_app);
	a_message->	FindRef("app_hint", & app_hint_ref);
	
	BMimeType mime	(type.String());

	if (pref_app == "none")
	{
		mime.SetPreferredApp(NULL);
		mime.SetAppHint(NULL);
	}
	else
		mime.SetPreferredApp(pref_app.String());
	
	app_hint_entry.SetTo(& app_hint_ref);

	if (app_hint_entry.Exists())
		mime.SetAppHint(& app_hint_ref);
}

void 
URL_Window::EmptyAppDropdown(void)
{
	while (1) 
	{
		BMenuItem * item = m_pref_app_menu->RemoveItem((int32)0);

		if (item == NULL)
			break;
		else
			delete item;
	}
}

void 
URL_Window::EmptyListview(void)
{
	while (1) 
	{
		BListItem * item = m_listview->RemoveItem((int32)0);

		if (item == NULL)
			break;
		else
			delete item;
	}
}

void 
URL_Window::DisplayDescriptions(void)
{
	BString * type = NULL;
	type = GetCurrentType();
	
	if (type->FindFirst("none") >= 0)
		m_longdescription->SetText("(no description)");	
	else
	{
		char *	buffer	=	new char [B_MIME_TYPE_LENGTH];
		BMimeType mime (type->String());
		mime.GetLongDescription(buffer);
		BString description(buffer);
				
		if (description.Length() <= 0)
			m_longdescription->SetText("(no description)");	
		else
			m_longdescription->SetText(description.String());	
			
		delete [] buffer;
	}

	delete type;	
}

void 
URL_Window::CenterWindowOnScreen(void)
{
	BScreen screen(this);
	BRect screen_size (screen.Frame());
	
	float new_left	=	(screen_size.right / 2);
	float new_top	=	(screen_size.bottom / 2);

	new_left -=	(Bounds().Width() / 2);
	new_top -=	(Bounds().Height() / 2);
	
	MoveTo(new_left, new_top);
}

bool 
URL_Window::IsWindowOffScreen(void)
{
	BScreen screen(this);
	BRect screen_frame (screen.Frame());
	BRect window_frame (Frame());
	
	if (window_frame.right	> screen_frame.right)	return true;
	if (window_frame.left	< screen_frame.left)	return true;
	if (window_frame.top	< screen_frame.top)		return true;
		// window tab may still be offscreen
	if (window_frame.bottom	> screen_frame.bottom)	return true;
		
	return false;
}

void 
URL_Window::PushWindowOnScreen(void)
{
	BScreen screen(this);
	BRect screen_frame (screen.Frame());
	BRect window_frame (Frame());
	
	float new_left	=	window_frame.left;
	float new_top	=	window_frame.top;
	
	if (window_frame.right > screen_frame.right)
		new_left = screen_frame.right - window_frame.Width();
	
	if (window_frame.left < screen_frame.left)
		new_left = 0;	

	if (window_frame.top < screen_frame.top)
		new_top = 40;
		
	if (window_frame.bottom > screen_frame.bottom)
		new_top = screen_frame.bottom - window_frame.Height();

	MoveTo(new_left, new_top);
}

